/*
 * Copyright 2009
 *     PYRASIS.COM,.  All rights reserved.
 *
 * Author:
 *     Lee Jae-Hong (pyrasis)
 */


#include <ntddk.h>
#include "condvar.h"


VOID
KeInitializeConditionVariable (
    __out PKCONDITION_VARIABLE ConditionVariable
    )
{
    KeInitializeEvent(&ConditionVariable->Event, NotificationEvent, FALSE);
    ConditionVariable->WakeAll = FALSE;
    ConditionVariable->Wake = FALSE;
    ConditionVariable->Waiters = 0;
}

NTSTATUS
KeSleepConditionVariableMX (
    __inout PKCONDITION_VARIABLE ConditionVariable,
    __inout PKMUTEX Mutex,
    __in_opt PLARGE_INTEGER Timeout
    )
{
    NTSTATUS status;
    KIRQL irql = KeGetCurrentIrql();
    ULONG order;

    //  
    order = ConditionVariable->Waiters;

    while (1)
    {
        InterlockedIncrement(&ConditionVariable->Waiters);

        KeReleaseMutex(Mutex, TRUE);
        status = KeWaitForSingleObject(&ConditionVariable->Event, Executive, KernelMode, FALSE, Timeout);

        KeWaitForMutexObject(Mutex, Executive, KernelMode, FALSE, NULL);

        InterlockedDecrement(&ConditionVariable->Waiters);

        if (status == STATUS_TIMEOUT)
        {
            return status;
        }

        if (ConditionVariable->WakeAll)
        {
            // ϰ ִ ü  ̻   ̺Ʈ Ѵ.
            if (ConditionVariable->Waiters == 0)
            {
                ConditionVariable->WakeAll = FALSE;
                KeResetEvent(&ConditionVariable->Event);
            }

            //  ϰ ִ ü   ̺Ʈ  ʰ
            // .
            break;
        }
        else if (ConditionVariable->Wake)
        {
            //    ü .
            if (order == 0)
            {
                ConditionVariable->Wake = FALSE;
                KeResetEvent(&ConditionVariable->Event);

                break;
            }
        }
        //  0 ƴ    ʰ ڽ  ø.
        order--;
    }

    return status;
}

NTSTATUS
KeSleepConditionVariableRS (
    __inout PKCONDITION_VARIABLE ConditionVariable,
    __inout PERESOURCE Resource,
    __in_opt PLARGE_INTEGER Timeout,
    __in ULONG Flags
    )
{
    NTSTATUS status;
    KIRQL irql = KeGetCurrentIrql();
    ULONG order;

    //  
    order = ConditionVariable->Waiters;

    while (1)
    {
        InterlockedIncrement(&ConditionVariable->Waiters);

        ExReleaseResourceLite(Resource);
        status = KeWaitForSingleObject(&ConditionVariable->Event, Executive, KernelMode, FALSE, Timeout);

        if (Flags == CONDITION_VARIABLE_LOCKMODE_SHARED)
            ExAcquireResourceSharedLite(Resource, TRUE);
        else
            ExAcquireResourceExclusiveLite(Resource, TRUE);

        InterlockedDecrement(&ConditionVariable->Waiters);

        if (status == STATUS_TIMEOUT)
        {
            return status;
        }

        if (ConditionVariable->WakeAll)
        {
            // ϰ ִ ü  ̻   ̺Ʈ Ѵ.
            if (ConditionVariable->Waiters == 0)
            {
                ConditionVariable->WakeAll = FALSE;
                KeResetEvent(&ConditionVariable->Event);
            }

            //  ϰ ִ ü   ̺Ʈ  ʰ
            // .
            break;
        }
        else if (ConditionVariable->Wake)
        {
            //    常 .
            if (order == 0)
            {
                ConditionVariable->Wake = FALSE;
                KeResetEvent(&ConditionVariable->Event);

                break;
            }
        }
        //  0 ƴ    ʰ ڽ  ø.
        order--;
    }

    return status;
}

VOID
KeWakeConditionVariable (
    __inout PKCONDITION_VARIABLE ConditionVariable
    )
{
    if (ConditionVariable->Waiters > 0)
    {
        ConditionVariable->Wake = TRUE;
        KeSetEvent(&ConditionVariable->Event, IO_NO_INCREMENT, FALSE);
    }
}

VOID
KeWakeAllConditionVariable (
    __inout PKCONDITION_VARIABLE ConditionVariable
    )
{
    if (ConditionVariable->Waiters > 0)
    {
        ConditionVariable->WakeAll = TRUE;
        KeSetEvent(&ConditionVariable->Event, IO_NO_INCREMENT, FALSE);
    }
}
